<?php

namespace Drupal\stackla_widget\Api;

use Drupal\Core\Config\ImmutableConfig;

/**
 * The Credentials class.
 *
 * @package Stackla\Core
 *
 * @property string $host
 * @property string $stack
 * @property string $key
 */
class Credentials {

  /**
   * Credentials Types.
   */
  const TYPE_OAUTH2 = 'oauth2';
  const TYPE_APIKEY = 'app_key';

  /**
   * Stackla domain name.
   *
   * @var string
   */
  public $host;

  /**
   * Stackla stack name.
   *
   * @var string
   */
  public $stack;

  /**
   * Authentication type.
   *
   * @var string
   */
  public $type;

  /**
   * OAuth2 Access token / normal API key.
   *
   * @var string
   */
  public $token;

  /**
   * OAuth2 refresh token.
   *
   * @var string
   */
  public $refreshToken;

  /**
   * OAuth2 type of token.
   *
   * @var string
   */
  public $tokenType;

  /**
   * OAuth2 access token expiring time.
   *
   * @var int
   */
  public $expiredIn;

  /**
   * OAuth2 state.
   *
   * @var string
   */
  private $state = "stackla-php-sdk";

  /**
   * Api key.
   *
   * @var string
   */
  private $apiKey;

  /**
   * The config factory service.
   *
   * @var \Drupal\Core\Config\ImmutableConfig
   */
  public $config;

  /**
   * The Constructor.
   *
   * @param \Drupal\Core\Config\ImmutableConfig $config_factory
   *   The config factory.
   * @param string $host
   *   API Host.
   * @param string $token
   *   API OAuth2 access token or normal Stackla API.
   * @param string $stack
   *   API stack.
   * @param string $type
   *   Authentication type, either TYPE_OAUTH2 or TYPE_APIKEY.
   */
  public function __construct(ImmutableConfig $config_factory, string $host = '', string $token = '', string $stack = '', string $type = Credentials::TYPE_OAUTH2) {
    $this->host = $host;
    $this->token = $token;
    $this->stack = $stack;
    $this->type = $type;
    $this->config = $config_factory;
  }

  /**
   * Get the host.
   *
   * @return string
   *   The host.
   */
  public function getHost() {
    return $this->host;
  }

  /**
   * Get the api key.
   *
   * @return string
   *   The api key.
   */
  public function getKey() {
    return $this->apiKey;
  }

  /**
   * Get the token.
   *
   * @return string
   *   The token.
   */
  public function getToken() {
    return $this->token;
  }

  /**
   * Get the stack.
   *
   * @return string
   *   The stack.
   */
  public function getStack() {
    return $this->stack;
  }

  /**
   * Set the host.
   *
   * @param string $host
   *   The host.
   */
  public function setHost($host) {
    $this->host = $host;
  }

  /**
   * Set the api key.
   *
   * @param string $apiKey
   *   The api key.
   */
  public function setKey($apiKey) {
    $this->apiKey = $apiKey;
  }

  /**
   * Set the token.
   *
   * @param string $token
   *   The token.
   */
  public function setToken($token) {
    $this->token = $token;
  }

  /**
   * Set the stack.
   *
   * @param string $stack
   *   The stack.
   */
  public function setStack($stack) {
    $this->stack = $stack;
  }

  /**
   * Generate access_token.
   *
   * @param string $client_id
   *   The client id.
   * @param string $client_secret
   *   The client secret.
   * @param string $access_code
   *   The access code.
   * @param string $redirect_uri
   *   The redirect uri.
   *
   * @return $this
   *
   * @throws \Exception
   */
  public function generateToken($client_id, $client_secret, $access_code, $redirect_uri) {
    $endpoint = "oauth2/token";
    $request = new Request($this, $this->host, $this->stack);
    $data = [
      'client_id' => $client_id,
      'client_secret' => $client_secret,
      'grant_type' => 'authorization_code',
      'code' => $access_code,
      'redirect_uri' => $redirect_uri,
    ];
    $json = $request->sendPost($endpoint, $data);
    $response = json_decode($json, TRUE);
    if (isset($response['error']) && count($response['error'])) {
      throw new \Exception("Error occurred: " .
        $response['error_description']);
    }
    $this->token = $response['access_token'];
    $this->tokenType = $response['token_type'];
    $this->expiredIn = $response['expires_in'];

    return $this;
  }

  /**
   * Generate access uri.
   *
   * @param string $client_id
   *   The client secret.
   * @param string $redirect_uri
   *   The redirect uri.
   *
   * @return string
   *   The access uri.
   */
  public function getAccessUri($client_id, $redirect_uri) {
    $endpoint = "oauth2/authenticate";

    $query = [
      'response_type' => 'code',
      'client_id' => $client_id,
      'redirect_uri' => $redirect_uri,
      'state' => $this->state,
    ];

    return sprintf("%s%s?%s", $this->host, $endpoint, http_build_query($query));
  }

}
