<?php

namespace Drupal\stackla_widget\Api\Exception;

/**
 * The Api Exception.
 *
 * @package Drupal\stackla_widget\Api\Exception
 */
class ApiException extends \Exception {

  /**
   * The HTTP status code.
   *
   * @var int
   */
  protected $httpStatusCode;

  /**
   * The HTTP reponse body.
   *
   * @var string
   */
  protected $httpResponseBody;

  /**
   * The error.
   *
   * @var ApiError[]
   */
  protected $errors;

  /**
   * {@inheritdoc}
   */
  public static function create($message, $httpStatusCode, $httpResponseBody, Exception $previous = NULL) {
    $exception = new static($message, 0, $previous);
    $exception->httpStatusCode = $httpStatusCode;
    $exception->httpResponseBody = $httpResponseBody;

    return $exception;
  }

  /**
   * Get HTTP status code.
   *
   * @return int
   *   The status code.
   */
  public function getHttpStatusCode() {
    return $this->httpStatusCode;
  }

  /**
   * Get HTTP response body.
   *
   * @return string
   *   The http response body.
   */
  public function getHttpResponseBody() {
    return $this->httpResponseBody;
  }

  /**
   * Get errors.
   *
   * @return ApiError[]
   *   The errors.
   */
  public function getErrors() {
    if (!is_array($this->errors)) {
      $errors = [];
      $json = $this->getHttpResponseBody();
      $response = json_decode($json, TRUE);
      if (json_last_error() === JSON_ERROR_NONE) {
        $errors = $this->ifx($response, 'errors');
        if (is_array($errors)) {
          foreach ($errors as $error) {
            if (is_array($error)) {
              $errors[] = new ApiError(
                  (string) $this->ifx($error, 'message'),
                  (string) $this->ifx($error, 'code'),
                  (string) $this->ifx($error, 'message_id'),
                  (int) $this->ifx($error, 'error_code')
              );
            }
          }
        }
      }
      $this->errors = $errors;
    }

    return $this->errors;
  }

  /**
   * Check error by code.
   *
   * @param int $errorCode
   *   The error code.
   *
   * @return bool
   *   Check status.
   */
  public function containsErrorByErrorCode($errorCode) {
    $errors = $this->getErrors();
    foreach ($errors as $error) {
      if ($error->getErrorCode() === $errorCode) {
        return TRUE;
      }
    }

    return FALSE;
  }

  /**
   * Check error by message id.
   *
   * @param string $messageId
   *   The message id.
   *
   * @return bool
   *   Check status.
   */
  public function containsErrorByMessageId($messageId) {
    $errors = $this->getErrors();
    foreach ($errors as $error) {
      if ($error->getMessageId() === $messageId) {
        return TRUE;
      }
    }

    return FALSE;
  }

  /**
   * Check if key exists.
   *
   * @param array $arr
   *   The array.
   * @param string $key
   *   The key.
   * @param array|null $default
   *   The default.
   *
   * @return mixed|null
   *   The result value.
   */
  final protected function ifx(array $arr = [], string $key = '', array $default = NULL) {
    if (!is_array($arr)) {
      return $default;
    }
    elseif (array_key_exists($key, $arr)) {
      return $arr[$key];
    }
    else {
      return $default;
    }
  }

}
