<?php

namespace Drupal\stackla_widget\Api;

use GuzzleHttp\Client;
use Monolog\Handler\StreamHandler;
use Monolog\Logger;
use Drupal\stackla_widget\Api\Exception\ApiException;
use Drupal\stackla_widget\Api\Validation\JsonValidator;

/**
 * The Request Class.
 *
 * @package Drupal\stackla_widget\Api
 */
class Request implements RequestInterface {

  /**
   * Stackla domain name.
   *
   * @var string
   *   The host.
   */
  protected $host;

  /**
   * Stackla stack name.
   *
   * @var string
   *   The stack.
   */
  protected $stack;

  /**
   * Stackla credentials.
   *
   * @var \Drupal\stackla_widget\Api\Credentials
   *   The credentials.
   */
  protected $credentials;

  /**
   * Response result placeholder.
   *
   * @var \GuzzleHttp\Message\Response
   *   The response http.
   */
  protected $response;

  /**
   * Request placeholder.
   *
   * @var \GuzzleHttp\Message\Request
   *   The request http.
   */
  protected $request;

  /**
   * Log.
   *
   * @var \Monolog\Logger
   *   The logger.
   */
  protected $logger = NULL;

  /**
   * The query seperator.
   *
   * @var string
   *   The seperator.
   */
  private $querySeparator = '&';

  /**
   * The api key.
   *
   * @var string
   *   The api key.
   */
  protected $apiKey;

  /**
   * The client.
   *
   * @var \GuzzleHttp\Message\Response
   *   The response.
   */
  private $client;

  /**
   * The Request constructor.
   *
   * @param Credentials $credentials
   *   The credentials object.
   * @param string $host
   *   The host.
   * @param string $stack
   *   The stack.
   */
  public function __construct(Credentials $credentials, string $host = '', string $stack = '') {
    $options = [];
    $configs = $credentials->config;
    if ($configs->get('proxy_status')) {
      $options = [
        'verify' => FALSE,
        'proxy' => [
          'http' => $configs->get('proxy_url'),
          'https' => $configs->get('proxy_url'),
        ],
      ];
    }
    $this->host = $host;
    $this->stack = $stack;
    $this->credentials = $credentials;
    // We prevent exception for being catched by the guzzle client.
    $this->client = new Client($options);

    if (class_exists("\\Monolog\\Logger")) {
      $this->logger = new Logger(get_class($this));
      $this->logger->pushHandler(new StreamHandler(sys_get_temp_dir() . DIRECTORY_SEPARATOR . 'stackla-request.log', Logger::INFO));
    }
  }

  /**
   * Set the host.
   *
   * @param string $host
   *   API host name.
   */
  public function setHost(string $host = '') {
    $this->host = $host;
  }

  /**
   * Set the api key.
   *
   * @param string $apiKey
   *   Client API key.
   */
  public function setApiKey(string $apiKey = '') {
    $this->apiKey = $apiKey;
  }

  /**
   * Set the stack.
   *
   * @param string $stack
   *   Client stack name.
   */
  public function setStack(string $stack = '') {
    $this->stack = $stack;
  }

  /**
   * Get the host.
   *
   * {@inheritdoc}
   */
  public function getHost() {
    return $this->host;
  }

  /**
   * Get the api key.
   *
   * {@inheritdoc}
   */
  public function getApiKey() {
    return $this->apiKey;
  }

  /**
   * Get the stack.
   *
   * {@inheritdoc}
   */
  public function getStack() {
    return $this->stack;
  }

  /**
   * Build a complete URI request by concanating name, endpoint, API key and id.
   *
   * @param string $endpoint
   *   Endpoint of API call.
   * @param array $query
   *   Query of data.
   *
   * @return string
   *   Complete URI.
   */
  private function buildUri(string $endpoint = '', array $query = []) {
    if ($this->credentials->type === Credentials::TYPE_OAUTH2) {
      $query = array_merge([
        'access_token' => $this->credentials->getToken(),
        'stack' => $this->stack,
      ], $query);
    }
    else {
      $query = array_merge([
        'api_key' => $this->credentials->getToken(),
        'stack' => $this->stack,
      ], $query);
    }
    // Prevent http_build_query ignore empty value.
    $query = $this->preventValueBeenIgnore($query);

    return sprintf(
      "%s/%s%s%s",
      rtrim($this->host, '/'),
      trim($endpoint, '/'),
      strpos($endpoint, '?') === FALSE ? '?' : '&',
      http_build_query($query, '', $this->querySeparator)
    );
  }

  /**
   * Prevent value been igonre.
   *
   * @param array $data
   *   The data.
   *
   * @return array|mixed
   *   The data.
   */
  public function preventValueBeenIgnore(array $data = []) {
    foreach ($data as $key => $value) {
      if (empty($value)) {
        $data[$key] = '';
      }
      elseif (gettype($value) == 'array') {
        $data[$key] = $this->preventValueBeenIgnore($value);
      }
    }

    return $data;
  }

  /**
   * Making request using Guzzle.
   *
   * @param string $method
   *   Type of request, either POST, GET, PUT or DELETE.
   * @param string $endpoint
   *   The operation / task for API.
   * @param array $data
   *   The parameter need to be passed.
   * @param array $options
   *   The options like header, body, etc.
   *
   * @return string
   *   The response body.
   *
   * @throws GuzzleHttp\Exception
   */
  private function sendRequest(string $method = 'GET', string $endpoint = '', array $data = [], array $options = []) {
    $uri = $this->buildUri($endpoint);

    if ($method === "GET" || $method === "PUT") {
      $uri = $this->buildUri($endpoint, $data);
    }
    elseif (count($data)) {
      $options['form_params'] = $data;
    }
    $this->response = $this->client->request($method, $uri, $options);
    if ($this->response->getStatusCode() >= 400) {
      $bt = debug_backtrace();
      $caller = $bt[2];
      if (isset($caller['class']) && $caller['class'] === get_class(new StacklaModel())) {
        $json = (string) $this->response->getBody();
        if (JsonValidator::validate($json, TRUE)) {
          $content = json_decode($json, TRUE);
          if (isset($content['errors'])) {
            $caller['object']->fromArray($content);
          }
        }
      }
      if ($this->logger) {
        $this->logger->addError(
          '-> REQUEST [' . $this->request->getMethod() . ' - ' . $this->request->getUrl() . "]",
          [$this->request->getMethod() !== "GET" ? (string) $this->request->getBody() : ""]
        );
        $this->logger->addError(
          '<- RESPONSE [' . $this->response->getStatusCode() . ':' . $this->response->getReasonPhrase() . "]",
          [(string) $this->response->getBody()]
        );
      }
    }
    else {
      if ($this->logger) {
        $this->logger->addInfo(
          '-> REQUEST [' . $this->request->getMethod() . ' - ' . $this->request->getUrl() . "]",
          [$this->request->getMethod() !== "GET" ? (string) $this->request->getBody() : ""]
        );
        $this->logger->addInfo(
          '<- RESPONSE [' . $this->response->getStatusCode() . ':' . $this->response->getReasonPhrase() . "]",
          [$this->response->json()]
        );
      }
    }

    $statusCode = $this->response->getStatusCode();
    switch ($statusCode) {
      case 200:
        return (string) $this->response->getBody();

      case 400:
        throw ApiException::create(
          sprintf(
            "Server return %s error code. Bad request: The request could not be understood. %s",
            $this->response->getStatusCode(),
            (string) $this->response->getBody()
          ),
          $statusCode,
          (string) $this->response->getBody()
        );

      case 401:
        throw ApiException::create(
          sprintf(
            "Server return %s error code. Unauthorized: Authentication credentials invalid or not authorised to access resource",
            $this->response->getStatusCode()
          ),
          $statusCode,
          (string) $this->response->getBody()
        );

      case 403:
        throw ApiException::create(
          sprintf(
            "
                  Server return %s error code. Rate limit exceeded: Too many requests in the current time window",
            $this->response->getStatusCode()
          ),
          $statusCode,
          (string) $this->response->getBody()
        );

      case 404:
        throw ApiException::create(
          sprintf(
            "Server return %s error code. Invalid resource: Invalid resource specified or resource not found",
            $this->response->getStatusCode()
          ),
          $statusCode,
          (string) $this->response->getBody()
        );

      default:
        throw ApiException::create(
          sprintf(
            "Server return %s error code.Server error: An error on the server prohibited a successful response; please contact support. %s",
            $this->response->getStatusCode(),
            (string) $this->response->getBody()
          ),
          $statusCode,
          (string) $this->response->getBody()
        );
    }

  }

  /**
   * Send a query GET.
   *
   * @param string $endpoint
   *   The endpoint.
   * @param array $data
   *   The data.
   * @param array $options
   *   The options.
   *
   * @return \GuzzleHttp\Message\Response|string
   *   The response.
   */
  public function sendGet(string $endpoint = '', array $data = [], array $options = []) {
    return $this->sendRequest('GET', $endpoint, $data, $options);
  }

  /**
   * Send a query POST.
   *
   * @param string $endpoint
   *   The endpoint.
   * @param array $data
   *   The data.
   * @param array $options
   *   The options.
   *
   * @return \GuzzleHttp\Message\Response|string
   *   The response.
   */
  public function sendPost(string $endpoint = '', array $data = [], array $options = []) {
    return $this->sendRequest('POST', $endpoint, $data, $options);
  }

  /**
   * Send a query PUT.
   *
   * @param string $endpoint
   *   The endpoint.
   * @param array $data
   *   The data.
   * @param array $options
   *   The options.
   *
   * @return \GuzzleHttp\Message\Response|string
   *   The response.
   */
  public function sendPut(string $endpoint = '', array $data = [], array $options = []) {
    return $this->sendRequest('PUT', $endpoint, $data, $options);
  }

  /**
   * Send a query DELETE.
   *
   * @param string $endpoint
   *   The endpoint.
   * @param array $data
   *   The data.
   * @param array $options
   *   The options.
   *
   * @return \GuzzleHttp\Message\Response|string
   *   The response.
   */
  public function sendDelete(string $endpoint = '', array $data = [], array $options = []) {
    return $this->sendRequest('DELETE', $endpoint, $data, $options);
  }

  /**
   * Get the status code.
   *
   * @return false
   *   The status code.
   */
  public function status() {
    if ($this->response) {
      return $this->response->getStatusCode();
    }

    return FALSE;
  }

  /**
   * Get the request.
   *
   * @return \GuzzleHttp\Message\Request
   *   The request.
   */
  public function getRequest() {
    return $this->request;
  }

  /**
   * Get response.
   *
   * @return \GuzzleHttp\Message\Response
   *   The response.
   */
  public function getResponse() {
    return $this->response;
  }

}
