<?php

namespace Drupal\stackla_widget\Api;

use Drupal\stackla_widget\Api\Validation\JsonValidator;
use Drupal\stackla_widget\Api\Validation\ModelAccessorValidator;
use Symfony\Component\Validator\Validation;
use Drupal\stackla_widget\Api\Tools\ReflectionUtil;

/**
 * The StacklaModel Class.
 *
 * @package Drupal\stackla_widget\Api
 */
class StacklaModel implements \IteratorAggregate, \Countable {

  /**
   * Stackla configs contains credentials, host, and stack.
   *
   * @var array
   */
  protected $configs;

  /**
   * Stackla domain name.
   *
   * @var string
   */
  protected $host;

  /**
   * Stackla stack name.
   *
   * @var string
   */
  protected $stack;

  /**
   * Endpoints.
   *
   * @var string
   */
  protected $endpoint;

  /**
   * Object properties.
   *
   * @var array
   */
  protected $propMap = [];

  /**
   * Updated object properties.
   *
   * @var array
   */
  protected $propMapUpdated = [];

  /**
   * Indicator if the object is only has a placeholder data or not.
   *
   * @var bool
   */
  protected $isPlaceholder = FALSE;

  /**
   * Request object.
   *
   * @var \Stackla\Core\Request
   */
  protected $request;

  /**
   * Error logs from request.
   *
   * @var array
   */
  protected $errors = [];

  /**
   * Query result as array.
   *
   * @var array
   */
  protected $records = [];

  /**
   * Types.
   *
   * @var array
   */
  protected $typeKeywords = [
    NULL,
    'string',
    'str',
    'integer',
    'int',
    'float',
    'double',
    'array',
    'mixed',
    'boolean',
    'bool',
  ];

  /**
   * Flag to keep the properties value.
   *
   * @var bool
   */
  protected $noTracking = FALSE;

  /**
   * Credentials.
   *
   * @var Credentials
   */
  protected $credentials;

  /**
   * The Constructor.
   *
   * @param array $configs
   *   The configs.
   * @param string|array $data
   *   The data.
   * @param bool $fetch
   *   Do get request to populate the field / property.
   */
  public function __construct(array $configs = [], $data = NULL, bool $fetch = TRUE) {
    $this->configs = array_merge([
      'credentials' => NULL,
      'host' => '',
      'stack' => '',
    ], $configs);

    $this->host = $this->configs['host'];
    $this->stack = $this->configs['stack'];

    if ($this->configs['credentials']) {
      $this->credentials = $this->configs['credentials'];
    }

    switch (gettype($data)) {
      case "NULL":
        break;

      case "string":
      case "integer":
        if (JsonValidator::validate($data) && gettype($data) !== 'integer') {
          $this->fromJson($data);
        }
        elseif (!empty($data)) {
          if ($fetch) {
            $this->getById($data);
          }
          else {
            $this->fromArray(['id' => $data]);
            $this->isPlaceholder = TRUE;
          }
        }
        break;

      case "array":
        $this->fromArray($data);
        break;

      default:
    }
  }

  /**
   * Magic isSet Method.
   *
   * @param string $key
   *   The key.
   *
   * @return bool
   *   The result.
   */
  public function __isset(string $key = '') {
    return isset($this->propMap[$key]);
  }

  /**
   * Magic Get Method.
   *
   * @param string $key
   *   The key.
   *
   * @return mixed
   *   The result.
   */
  public function __get(string $key = '') {
    $camelCase = $this->convertToCamelCase($key);
    if ($this->__isset($key)) {
      return $this->propMap[$key];
    }
    elseif (ModelAccessorValidator::validate($this, $camelCase)) {
      $annots = ReflectionUtil::propertyAnnotations($this, $camelCase, 'set');
      if (isset($annots['uses'])) {
        $getter = "get{$camelCase}";
        return call_user_func([$this, $getter]);
      }

    }

    return NULL;
  }

  /**
   * Magic Set Method.
   *
   * @param string $key
   *   The key.
   * @param string $value
   *   The value.
   *
   * @throws \Exception
   */
  public function __set(string $key = '', $value = '') {
    $validKey = FALSE;
    $camelCase = $this->convertToCamelCase($key);
    $readonly = FALSE;
    $trace = debug_backtrace();

    $property = NULL;
    if (ModelAccessorValidator::property($this, lcfirst($camelCase))) {
      $property = lcfirst($camelCase);
    }

    if (ModelAccessorValidator::validate($this, $camelCase)) {
      $annots = ReflectionUtil::propertyAnnotations($this, $camelCase, 'set');
      $setter = "set{$camelCase}";
      $ref = new \ReflectionMethod($this, $setter);
      $permitedCaller = isset($trace[1]) && isset($trace[1]['function']) && $trace[1]['function'] === $setter;
      if (!$ref->isPublic() && !$permitedCaller) {
        $readonly = TRUE;
      }
      else {
        if ($property) {
          $this->$property = $value;
        }
        if (isset($trace[1]) && $trace[1]['function'] != $setter && isset($annots['uses'])) {
          call_user_func([$this, $setter], $value);
        }
        else {
          $validKey = TRUE;
        }
      }
    }
    elseif ($property) {
      $validKey = TRUE;

      $annotations = ReflectionUtil::mapper($this, $property);

      // Allow to assign value if it's comming from assignValue.
      $caller = 'assignValue';
      $permitedCaller = isset($trace[1]) && isset($trace[1]['function']) && $trace[1]['function'] === $caller;

      if (ReflectionUtil::propertyAccess($this, $key) === 'read' && !$permitedCaller) {
        $readonly = TRUE;
      }
      else {
        $clazz = isset($annotations['var']) ? $annotations['var'] : NULL;
        $clazz = preg_replace("/(\\[\\]|\\(\\))$/", "", $clazz);
        if (!empty($value) && !in_array($clazz, $this->typeKeywords)) {
          $o = new $clazz();
          $setter = sprintf("set%s", $camelCase);
          if (method_exists($o, $setter)) {
            $o->$setter($value);
            $value = $o;
          }
          elseif (method_exists($o, 'formArray')) {
            $o->fromArray($value);
            $value = $o;
          }
        }
        $this->$property = $value;
      }
    }

    // Throw an error if user try to override/assign the value.
    if ($readonly) {
      throw new \Exception("Attempted to write readonly property '{$key}' of " . get_class($this) . " class.");
    }

    // Assign value to property mapper.
    if ($validKey) {
      $old_value = $this->__isset($key) ? $this->__get($key) : NULL;
      if ($value === NULL) {
        $this->__unset($key);
      }
      else {
        if (!$this->noTracking && $old_value !== $value) {
          $this->propMapUpdated[] = $key;
        }
        $this->propMap[$key] = $value;
      }
    }
  }

  /**
   * Magic Unset Method.
   *
   * @param string $key
   *   The key.
   */
  public function __unset(string $key = '') {
    unset($this->propMap[$key]);
  }

  /**
   * Mapping Annotations.
   *
   * @param string $class
   *   The class.
   * @param string $property
   *   The property.
   *
   * @return array|null
   *   The result.
   */
  protected function mapAnnotations(string $class = '', string $property = '') {
    $ref = new \ReflectionProperty($class, $property);
    if (!preg_match_all('~\@([^\s@\(]+)[\t ]*(?:\(?([^\n@]+)\)?)?~i', $ref->getDocComment(), $matchAnnots, PREG_PATTERN_ORDER)) {
      return NULL;
    }

    $annotations = [];
    foreach ($matchAnnots[1] as $i => $annot) {
      $annotations[$annot] = empty($matchAnnots[2][$i]) ? TRUE : rtrim($matchAnnots[2][$i], " \t\n\r");
    }

    return $annotations;
  }

  /**
   * Check if property is public or not.
   *
   * @param string $key
   *   Property name.
   *
   * @return bool
   *   The result.
   */
  public function isPublic(string $key = '') {
    $camelCase = $this->convertToCamelCase($key);
    $setter = "set{$camelCase}";

    // If class doesn't have setter method, we threat it as public property.
    if (!method_exists($this, $setter)) {
      return TRUE;
    }

    $ref = new \ReflectionMethod($this, $setter);
    if ($ref->isPublic()) {
      return TRUE;
    }

    return FALSE;
  }

  /**
   * Reset properties tracking for update perpose.
   *
   * @return $this
   *   The result.
   */
  protected function resetTracking() {
    $this->propMapUpdated = [];

    return $this;
  }

  /**
   * Check if property is been updated or not.
   *
   * @param string $property
   *   The property.
   *
   * @return bool
   *   The result.
   */
  protected function isUpdated(string $property = '') {
    // Check if property is been updated.
    // Check if property is public.
    // Check if property name start with '_'.
    if (in_array($property, $this->propMapUpdated) && $this->isPublic($property) && !preg_match("/^_/", $property)) {
      return TRUE;
    }

    return FALSE;
  }

  /**
   * Converts Params to Array.
   *
   * @param array $param
   *   The param.
   * @param bool $only_updated
   *   The only updated.
   *
   * @return array
   *   The result.
   */
  protected function convertToArray(array $param = [], bool $only_updated = FALSE) {
    $ret = [];
    foreach ($param as $k => $v) {
      // Exclude protected properties to be exported to array.
      if ($only_updated && !$this->isUpdated($k)) {
        continue;
      }

      if ($v instanceof StacklaModel) {
        $ret[$k] = $v->toArray($only_updated);
      }
      else {
        $ret[$k] = $v;
      }
    }
    // If the array is empty, which means an empty object.
    // We need to convert array to StdClass object to properly.
    // Represent JSON String.
    if (count($ret) <= 0) {
      $ret = [];
    }
    return $ret;
  }

  /**
   * Returns array representation of object.
   *
   * @param bool $only_updated
   *   Boolean value.
   *
   * @return array
   *   The result.
   */
  public function toArray(bool $only_updated = FALSE) {
    return $this->convertToArray($this->propMap, $only_updated);
  }

  /**
   * Converts the input key into a valid Setter Method Name.
   *
   * @param string $key
   *   The key.
   *
   * @return mixed
   *   The result.
   */
  private function convertToCamelCase(string $key = '') {
    return str_replace(' ', '', ucwords(str_replace(['_', '-'], ' ', $key)));
  }

  /**
   * Converts the input key into a valid non Setter Method Name.
   *
   * @param string $key
   *   The key.
   *
   * @return mixed
   *   The result.
   */
  private function convertToNonCamelCase(string $key = '') {
    return strtolower(preg_replace('/(?<=\\w)(?=[A-Z])/', "_$1", $key));
  }

  /**
   * Fills object value from Array list.
   *
   * @param array $arr
   *   Array of object data.
   *
   * @return $this
   *   The result.
   *
   * @throws \Exception
   */
  public function fromArray(array $arr = []) {
    // don't track the changes for fields.
    $this->resetTracking();
    $this->noTracking = TRUE;

    if (!empty($arr)) {
      if (isset($arr['data']) && isset($arr['errors'])) {
        $this->setErrors($arr['errors']);
        $this->setData($arr['data']);
      }
      else {
        // Iterate over each element in array.
        foreach ($arr as $k => $v) {
          // If the value is an array, it is an object after conversion.
          if (is_array($v) && !empty($v)) {
            // Determine the class of the object.
            if (($clazz = ReflectionUtil::getPropertyClass(get_class($this), $k)) != NULL && !in_array($clazz, $this->typeKeywords)) {
              // If the value is an associative array, its an object.
              if ($this->isAssocArray($v)) {
                $o = new $clazz();
                $o->fromArray($v);
                $this->assignValue($k, $o);
              }
              else {
                // Else, value is an array of object/data.
                $arr = [];
                // Iterate through each element in that array.
                foreach ($v as $nk => $nv) {
                  if (is_array($nv)) {
                    $o = new $clazz();
                    $o->fromArray($nv);
                    $arr[$nk] = $o;
                  }
                  else {
                    $arr[$nk] = $nv;
                  }
                }
                $this->assignValue($k, $arr);
              }
            }
            else {
              $this->assignValue($k, $v);
            }
          }
          else {
            $this->assignValue($k, $v);
          }
        }
      }
    }
    elseif ($arr === FALSE) {
      $reqRes = $this->request ? $this->request->getResponse() : NULL;
      if ($reqRes) {
        $responseBody = $reqRes->getBody(TRUE);
        $msg = [];
        if ($responseBody != '') {
          $response = json_decode($responseBody, TRUE);
          if (count($response['errors'])) {
            foreach ($response['errors'] as $error) {
              $msg[] = isset($error['message']) ? "[" . $error['code'] . "] " . $error['message'] : json_encode($error);
            }
          }
        }
        $this->setErrors($msg);
        throw new \Exception(implode("\r\n", $msg), $reqRes->getStatusCode());
      }
      else {
        throw new \Exception('Error while adding data to ' . get_class($this) . ' object');
      }

    }
    $this->noTracking = FALSE;
    $this->isPlaceholder = FALSE;

    return $this;
  }

  /**
   * Assign value.
   *
   * @param string $key
   *   The key.
   * @param string $value
   *   The value.
   *
   * @throws \Exception
   */
  private function assignValue(string $key = '', $value = '') {
    // If we find the getter setter, use that, otherwise use magic method.
    if (ModelAccessorValidator::validate($this, $this->convertToCamelCase($key))) {
      $setter = "set" . $this->convertToCamelCase($key);
      $this->$setter($value);
    }
    else {
      $this->__set($key, $value);
    }
  }

  /**
   * Set data.
   *
   * @param array $data
   *   The data.
   *
   * @throws \Exception
   */
  private function setData(array $data = []) {
    if (!$this->isAssocArray($data)) {
      $class = get_class($this);

      foreach ($data as $item) {
        $this->records[] = new $class($this->configs, $item, FALSE);
      }
    }
    elseif (!empty($data)) {
      $this->records[] = $this->fromArray($data);
    }
  }

  /**
   * Request error.
   *
   * @param array $errors
   *   Errors.
   *
   * @return $this
   *   The result.
   */
  private function setErrors(array $errors = []) {
    $this->errors = $errors;
    return $this;
  }

  /**
   * Get errors.
   *
   * @return array
   *   List errors.
   */
  public function getErrors() {
    return $this->errors;
  }

  /**
   * Fills object value from Json string/object.
   *
   * @param string $json
   *   Json data.
   *
   * @return $this
   *   The result.
   */
  public function fromJson(string $json = '') {
    if (is_string($json)) {
      $json = json_decode($json, TRUE);
    }

    return $this->fromArray($json);
  }

  /**
   * Init Request.
   *
   * @param bool $force
   *   Force.
   */
  protected function initRequest(bool $force = FALSE) {
    if (!$this->request || $force) {
      $this->request = new Request($this->credentials, $this->host, $this->stack);
    }
  }

  /**
   * API credentials setter.
   *
   * @param \Stackla\Core\Credentials $credentials
   *   Credentials object.
   *
   * @return $this
   *   The result.
   */
  public function setCredentials(Credentials $credentials) {
    $this->credentials = $credentials;

    return $this;
  }

  /**
   * API Credentials getter.
   *
   * @return Credentials
   *   Credentials object.
   */
  public function getCredentials() {
    return $this->credentials;
  }

  /**
   * This method will create new content in Stackla.
   *
   * @return mixed
   *   Will return FALSE or json object.
   */
  public function create() {
    $endpoint = sprintf("%s", $this->endpoint);

    $this->initRequest();

    $json = $this->request->sendPost($endpoint, $this->toArray());

    $this->fromJson($json);

    return $json === FALSE ? FALSE : $this;
  }

  /**
   * This method will return content.
   *
   * @return mixed
   *   Will return FALSE or json object.
   */
  public function get() {
    $arg_list = func_get_args();

    $endpoint = sprintf("%s", $this->endpoint);

    $options = isset($arg_list[2]) ? $arg_list[2] : [];

    $data = array_merge(
      [
        'limit' => isset($arg_list[0]) ? $arg_list[0] : 500,
        'page' => isset($arg_list[1]) ? $arg_list[1] : 1,
      ],
      $options
    );

    $this->initRequest();

    $json = $this->request->sendGet($endpoint, $data);

    $this->fromJson($json);
    return $json === FALSE ? FALSE : $this;
  }

  /**
   * This method will return content by provided valid ID.
   *
   * @param int|string $id
   *   Content id.
   *
   * @return mixed
   *   Will return FALSE or json object.
   */
  public function getById($id) {
    $endpoint = sprintf("%s/%s", $this->endpoint, $id);

    $this->initRequest();

    $json = $this->request->sendGet($endpoint);

    $this->fromJson($json);
    return $json === FALSE ? FALSE : $this;
  }

  /**
   * This method will update content in Stackla.
   *
   * @param bool $force
   *   Indicator to force update even the object is using placeholder.
   *
   * @return mixed
   *   Will return FALSE or json object.
   *
   * @throws \Exception
   */
  public function update(bool $force = FALSE) {
    if ($this->isPlaceholder && !$force) {
      throw new \Exception("This is placeholder object, it doesn't have a uptodate data.
      If you still want to update this object with provided property(ies),
      you can pass 'true' value to the first parameter of this method");
    }

    $endpoint = sprintf("%s/%s", $this->endpoint, $this->id);

    $this->initRequest();

    $data = $this->toArray(TRUE);
    unset($data['id']);

    $json = $this->request->sendPut($endpoint, $data);

    $this->fromJson($json);
    return $json === FALSE ? FALSE : $this;
  }

  /**
   * This method will delete content from Stackla.
   *
   * @return mixed
   *   Will return FALSE or json object.
   */
  protected function delete() {
    $endpoint = sprintf("%s/%s", $this->endpoint, $this->id);

    $this->initRequest();

    $json = $this->request->sendDelete($endpoint);

    $this->fromJson($json);
    return $json === FALSE ? FALSE : $this;
  }

  /**
   * Get validations.
   *
   * @return array
   *   The result.
   */
  public function getValidations() {
    $validations = [];
    $validator = Validation::createValidatorBuilder()
      ->enableAnnotationMapping()
      ->getValidator();
    $class = get_class($this);
    $metadata = $validator->getMetadataFor(new $class());
    $constrainedProperties = $metadata->getConstrainedProperties();

    // Loop for all properties' that has constraints / rules.
    foreach ($constrainedProperties as $constrainedProperty) {
      $propertyMetadata = $metadata->getPropertyMetadata($constrainedProperty);
      $constraints = $propertyMetadata[0]->constraints;
      $outputConstraintsCollection = [];

      // Loop for all constraints / rules.
      foreach ($constraints as $constraint) {
        $class = new \ReflectionObject($constraint);
        $constraintName = $class->getShortName();
        switch ($constraintName) {
          case "NotBlank":
            $param = "notBlank";
            break;

          case "Type":
            $param = $constraint->type;
            break;

          case "Choice":
            $param = $constraint->choices;
            break;

          case "Url":
            $param = $constraint->protocols;
            break;

          default:
            $param = $constraint;
            break;
        }

        $outputConstraintsCollection[$constraintName] = $param;
      }
      $sourceProp = preg_replace('/^_/', '', $this->convertToNonCamelCase($constrainedProperty));

      $validations[$sourceProp] = $outputConstraintsCollection;
    }

    return $validations;
  }

  /**
   * Validate.
   *
   * @return array
   *   The result.
   */
  public function validate() {
    $validations = [];

    $validator = Validation::createValidatorBuilder()
      ->enableAnnotationMapping()
      ->getValidator();

    $validations = $validator->validate($this);

    foreach ($validations as $validation) {
      $prop = preg_replace('/^_/', '', $validation->getPropertyPath());
      $sourceProp = $this->convertToNonCamelCase($prop);

      $prop = $sourceProp;
      $validations[] = [
        'property' => $prop,
        'message' => $validation->getMessage(),
      ];
    }

    return $validations;
  }

  /**
   * Get response code.
   *
   * @return mixed
   *   The result.
   */
  public function getResponseCode() {
    return $this->request->status();
  }

  /**
   * Get the total number of item.
   *
   * @return int
   *   The count.
   */
  public function count() {
    return count($this->records);
  }

  /**
   * Returns all results.
   *
   * @return array
   *   List of records.
   */
  public function getResults() {
    return $this->records;
  }

  /**
   * Returns the items for iteration.
   *
   * @return \ArrayIterator
   *   The result.
   */
  public function getIterator() {
    return new \ArrayIterator($this->records);
  }

  /**
   * Check assoc array.
   *
   * @param array $arr
   *   The array.
   *
   * @return true
   *   If $arr is an associative array.
   */
  public static function isAssocArray(array $arr = []) {
    foreach ($arr as $k => $v) {
      if (is_int($k)) {
        return FALSE;
      }
    }
    return TRUE;
  }

}
