<?php

namespace Drupal\stackla_widget\Api\Tools;

use Drupal\stackla_widget\Api\Exception\StacklaConfigurationException;
use Drupal\stackla_widget\Api\StacklaModel;

/**
 * The ReflectionUtil Class.
 *
 * @package Stackla\Core
 */
class ReflectionUtil {
  /**
   * Reflection Class.
   *
   * @var \ReflectionClass[]
   */
  private static $classRefl = [];

  /**
   * Class definition.
   *
   * @var array
   */
  private static $classDefs = [];

  /**
   * Reflection Methods.
   *
   * @var \ReflectionMethod[]
   */
  private static $methodsRefl = [];

  /**
   * Reflection Methods return type.
   *
   * @var \ReflectionMethod[]
   */
  private static $methodsType = [];

  /**
   * Reflection Properties.
   *
   * @var \ReflectionProperty[]
   */
  private static $propertiesRefl = [];

  /**
   * Properties Type.
   *
   * @var string[]
   */
  private static $propertiesType = [];

  /**
   * Gets Property Class of the given property.
   *
   * @param string $class
   *   The class.
   * @param string $propertyName
   *   The property name.
   *
   * @return null|string
   *   The result.
   *
   * @throws \Drupal\stackla_widget\Api\Exception\StacklaConfigurationException
   */
  public static function getPropertyClass(string $class = '', string $propertyName = '') {
    if ($class == get_class(new StacklaModel())) {
      // Make it generic if StacklaModel is used for generating this.
      return get_class(new StacklaModel());
    }

    // If the class doesn't exist, or the method doesn't exist, return NULL.
    if (!class_exists($class) || !method_exists($class, self::getter($class, $propertyName))) {
      return NULL;
    }

    if (($annotations = self::propertyAnnotations($class, $propertyName)) && isset($annotations['return'])) {
      $param = $annotations['return'];
    }

    if (isset($param)) {
      $anno = preg_split("/[\\s\\[\\]]+/", $param);
      return $anno[0];
    }
    else {
      throw new StacklaConfigurationException("Getter function for '$propertyName' in '$class' class should have a proper return type.");
    }
  }

  /**
   * Retrieves Annotations of each property.
   *
   * @param string $class
   *   The class.
   * @param string $propertyName
   *   The property name.
   * @param string $getset
   *   The get set.
   *
   * @throws \RuntimeException
   *
   * @return mixed
   *   The result.
   */
  public static function propertyAnnotations(string $class = '', string $propertyName = '', string $getset = 'get') {
    $class = is_object($class) ? get_class($class) : $class;
    if (!class_exists('ReflectionProperty')) {
      throw new \RuntimeException("Property type of " . $class . "::{$propertyName} cannot be resolved");
    }

    if ($annotations =& self::$methodsType[$class][$propertyName]) {
      return $annotations;
    }

    if (!($refl =& self::$methodsRefl[$class][$propertyName])) {
      if ($getset == 'set') {
        $method = self::setter($class, $propertyName);
      }
      else {
        $method = self::getter($class, $propertyName);
      }
      $refl = new \ReflectionMethod($class, $method);
      self::$methodsRefl[$class][$propertyName] = $refl;
    }

    $annots = self::parse($refl->getDocComment());

    if (!$annots) {
      return NULL;
    }

    foreach ($annots[1] as $i => $annot) {
      $annotations[strtolower($annot)] = empty($annots[2][$i]) ? TRUE : rtrim($annots[2][$i], " \t\n\r)");
    }

    return $annotations;
  }

  /**
   * The property Access.
   *
   * @param object $class
   *   The class.
   * @param string $property
   *   The property.
   *
   * @return mixed|null
   *   The result.
   *
   * @throws \ReflectionException
   */
  public static function propertyAccess($class = NULL, string $property = '') {
    $default = NULL;
    $class = is_object($class) ? get_class($class) : $class;
    if (!class_exists('ReflectionProperty')) {
      throw new \RuntimeException("Property type of " . $class . "::{$property} cannot be resolved");
    }

    if (!($refl =& self::$classRefl[$class][$property])) {
      $refl = new \ReflectionClass($class);
      self::$classRefl[$class][$property] = $refl;
    }

    if (!($defs =& self::$classDefs[$class])) {
      $annots = self::parse($refl->getDocComment());
      if (!$annots) {
        return NULL;
      }

      $annotations = [];
      foreach ($annots[3] as $i => $annot) {
        if (empty($annot)) {
          continue;
        }
        $annotations[$annot] = empty($annots[1][$i]) ? $default : preg_replace("~^property((-|)([\w]+|))~", "$3", rtrim($annots[1][$i], " \t\n\r"));
        if ($annotations[$annot] == "") {
          $annotations[$annot] = $default;
        }
      }
      $defs = $annotations;
    }
    if (isset($defs[$property])) {
      return $defs[$property];
    }
    return $default;
  }

  /**
   * The mapper function.
   *
   * @param object $class
   *   The class.
   * @param string $property
   *   The property.
   *
   * @return array|null
   *   The result.
   *
   * @throws \ReflectionException
   */
  public static function mapper($class = NULL, string $property = '') {
    $class = is_object($class) ? get_class($class) : $class;
    if (!class_exists('ReflectionProperty')) {
      throw new \RuntimeException("Property type of " . $class . "::{$property} cannot be resolved");
    }

    if (!($refl =& self::$propertiesRefl[$class][$property])) {
      $refl = new \ReflectionProperty($class, $property);
      self::$propertiesRefl[$class][$property] = $refl;
    }

    $annots = self::parse($refl->getDocComment());

    if (!$annots) {
      return NULL;
    }

    $annotations = [];
    foreach ($annots[1] as $i => $annot) {
      $annotations[$annot] = empty($annots[2][$i]) ? TRUE : rtrim($annots[2][$i], " \t\n\r");
    }

    return $annotations;
  }

  /**
   * The parse function.
   *
   * @param string $docComment
   *   The doc comment.
   *
   * @return null
   *   The result.
   */
  private static function parse(string $docComment = '') {
    if (!preg_match_all(
      '~\@([^\s@\(]+)[\t ]*(?:\(?([^\n\s@]+)\)?)?[\t ]*(\$(?:\(?([^\$\n\s@]+)\)?)?[\t ]*(?:\(?([^\n@]+)\)?))?~i',
      $docComment,
      $annots,
      PREG_PATTERN_ORDER)
    ) {
      return NULL;
    }
    return $annots;
  }

  /**
   * Preg_replace_callback callback function.
   *
   * @param string|array $match
   *   The match.
   *
   * @return string
   *   The result.
   */
  private static function replaceCallback($match = '') {
    return ucwords($match[2]);
  }

  /**
   * Returns the properly formatted getter function name.
   *
   * @param string $class
   *   The class.
   * @param string $propertyName
   *   The property name.
   *
   * @return string
   *   Getter function name.
   */
  public static function setter(string $class = '', string $propertyName = '') {
    return method_exists($class, "set" . ucfirst($propertyName)) ?
      "set" . ucfirst($propertyName) :
      "set" . preg_replace_callback("/([_\\-\\s]?([a-z0-9]+))/", "self::replaceCallback", $propertyName);
  }

  /**
   * Returns the properly formatted getter function name.
   *
   * @param string $class
   *   The class.
   * @param string $propertyName
   *   The property name.
   *
   * @return string
   *   Getter function name.
   */
  public static function getter(string $class = '', string $propertyName = '') {
    return method_exists($class, "get" . ucfirst($propertyName)) ?
      "get" . ucfirst($propertyName) :
      "get" . preg_replace_callback("/([_\\-\\s]?([a-z0-9]+))/", 'self::replaceCallback', $propertyName);
  }

}
