<?php

namespace Drupal\stackla_widget\Api;

/**
 * The Widget Class.
 *
 * Stakla\Api.
 *
 * @property array $style
 * @property array $config
 * @property string $css
 * @property int $filter_id
 * @property string $custom_css
 * @property string $custom_js
 * @property string $external_js
 * @property int $parent_id
 * @property-read string $embed_code
 * @property-read int $id
 * @property-read int $stackla_id
 * @property-read string $guid
 */
class Widget extends StacklaModel implements WidgetInterface {


  const TEXT_TILE_FONT_COLOR = '666666';
  const TEXT_TILE_USER_NAME_FONT_COLOR = '333333';
  const WIDGET_LOADING_IMAGE = '/media/images/widget/loading.gif';
  const CLICK_THROUGH_URL_1 = '%5BEXPAND%5D';
  const CLICK_THROUGH_URL_2 = '%5BORIGINAL_URL%5D';

  /**
   * Fluid type style group.
   *
   * @var array
   */
  public static $fluidTypeStyles = [
    "fluid",
    "horizontal-fluid",
    "base_waterfall",
    "base_carousel",
  ];
  /**
   * Static type style group.
   *
   * @var array
   */
  public static $staticTypeStyles = [
    "carousel",
    "main",
    "slideshow",
    "auto",
    "base_feed",
    "base_billboard",
    "base_slideshow",
  ];

  /**
   * Endpoints.
   *
   * @var string
   */
  protected $endpoint = 'widgets';

  /**
   * Unique identifier for the widget.
   *
   * @var int
   */
  protected $id;

  /**
   * Stack id.
   *
   * @var int
   */
  protected $stackId;

  /**
   * Widget guid.
   *
   * @var string
   */
  protected $guid;

  /**
   * Style.
   *
   * @var array
   * @Assert\NotBlank()
   * @Assert\NotNull()
   */
  protected $style;

  /**
   * Config.
   *
   * @var array
   */
  protected $config;

  /**
   * Css.
   *
   * @var string
   * @Assert\Type(type="string")
   */
  protected $css;

  /**
   * Filter id.
   *
   * @var int
   * @Assert\NotBlank()
   * @Assert\NotNull()
   * @Assert\Type(type="integer")
   */
  protected $filterId;

  /**
   * Status enabled.
   *
   * @var int
   */
  protected $enabled;

  /**
   * Custom css.
   *
   * @var string
   * @Assert\Type(type="string")
   */
  protected $customCss;

  /**
   * Source css.
   *
   * @var string
   * @Assert\Type(type="string")
   */
  protected $sourceCss;

  /**
   * Custom js.
   *
   * @var string
   * @Assert\Type(type="string")
   */
  protected $customJs;

  /**
   * External js.
   *
   * @var string
   * @Assert\Type(type="string")
   */
  protected $externalJs;

  /**
   * Parent id.
   *
   * @var int
   * @Assert\Type(type="integer")
   */
  protected $parentId;

  /**
   * Embed code.
   *
   * @var string
   * @Assert\Type(type="string")
   */
  protected $embedCode;

  /**
   * Widget name.
   *
   * @var string
   * @uses $style['name']
   * @Assert\NotBlank()
   * @Assert\NotNull()
   * @Assert\Type(type="string")
   * @Assert\Length(min=3, max=255)
   */
  protected $name;

  /**
   * Widget type.
   *
   * @var string
   * @uses $style['type']
   * @Assert\Choice(choices={"fluid", "fixed"})
   * @Assert\NotBlank()
   * @Assert\NotNull()
   * @Assert\Type(type="string")
   */
  protected $type;

  /**
   * Widget name.
   *
   * @var string
   * @uses $style['style']
   * @Assert\Choice(choices={"fluid", "horizontal-fluid", "carousel", "main", "slideshow",
   *  "auto", "base_waterfall", "base_carousel", "base_feed",
   *   "base_billboard", "base_slidehow"})
   * @Assert\NotBlank()
   * @Assert\NotNull()
   * @Assert\Type(type="string")
   */
  protected $typeStyle;

  /**
   * Widget name.
   *
   * @var int
   * @uses $style['max_tile_width']
   */
  protected $maxTileWidth;

  /**
   * Widget name.
   *
   * @param string $name
   *   The name.
   *
   * @return $this
   *
   * @uses $style['name']
   */
  public function setName($name) {
    $style = $this->style;
    if (empty($style)) {
      $style = [];
    }
    $style['name'] = $name;
    $this->style = $style;

    return $this;
  }

  /**
   * Widget name.
   *
   * @return string
   *   The result.
   */
  public function getName() {
    $styleValue = $this->style;

    if (!isset($styleValue['name'])) {
      return NULL;
    }
    else {
      return $styleValue['name'];
    }
  }

  /**
   * Widget type.
   *
   * @param string $type
   *   The type.
   *
   * @return $this
   *
   * @uses $style['type']
   */
  public function setType($type) {
    $style = $this->style;
    if (empty($style)) {
      $style = [];
    }
    $style['type'] = $type;
    $this->style = $style;

    return $this;
  }

  /**
   * Widget type.
   *
   * @return string
   *   The result.
   */
  public function getType() {
    $styleValue = $this->style;

    if (!isset($styleValue['type'])) {
      return NULL;
    }
    else {
      return $styleValue['type'];
    }
  }

  /**
   * Widget type style.
   *
   * @param string $typeStyle
   *   The type style.
   *
   * @return $this
   *
   * @uses $style['style']
   */
  public function setTypeStyle($typeStyle) {
    $style = $this->style;
    if (empty($style)) {
      $style = [];
    }
    $style['style'] = $typeStyle;
    $this->style = $style;

    if (in_array($typeStyle, self::$fluidTypeStyles)) {
      $this->type = self::TYPE_FLUID;
    }
    elseif (in_array($typeStyle, self::$staticTypeStyles)) {
      $this->type = self::TYPE_STATIC;
    }

    return $this;
  }

  /**
   * Widget type style.
   *
   * @return string
   *   The result.
   */
  public function getTypeStyle() {
    $styleValue = $this->style;

    if (!isset($styleValue['style'])) {
      return NULL;
    }
    else {
      return $styleValue['style'];
    }
  }

  /**
   * Widget max tile width.
   *
   * @param int $width
   *   The width.
   *
   * @return $this
   *
   * @uses $style['max_tile_width']
   */
  public function setMaxTileWidth($width) {
    $style = $this->style;
    if (empty($style)) {
      $style = [];
    }
    $style['max_tile_width'] = $width;
    $this->style = $style;

    return $this;
  }

  /**
   * Widget max tile width.
   *
   * @return int
   *   The result.
   */
  public function getMaxTileWidth() {
    $styleValue = $this->style;

    if (!isset($styleValue['max_tile_width'])) {
      return NULL;
    }
    else {
      return $styleValue['max_tile_width'];
    }
  }

  /**
   * Set default value.
   */
  public function setDefaultValue() {
    $defaultConfig = [
      'tile_options' => [
        'show_tags' => '0',
        'show_votes' => '0',
        'show_likes' => '0',
        'show_dislikes' => '0',
        'show_comments' => '0',
        'show_shopspots' => '0',
      ],
      'lightbox' => [
        'layout' => 'portrait',
        'show_additional_info' => '1',
        'show_sharing' => '0',
        'sharing_text' => '',
        'sharing_title' => '',
        'show_comments' => '0',
        'post_comments' => '0',
        'show_products' => '0',
        'show_shopspots' => '0',
      ],
    ];
    $style = $this->style;
    $config = $this->config;
    if (empty($style)) {
      $style = [];
    }
    if (empty($config)) {
      $config = ['tile_options' => [], 'lightbox' => []];
    }
    switch ($this->type_style) {
      case self::STYLE_BASE_WATERFALL:
      case self::STYLE_VERTICAL_FLUID:
        $defaultStyle = [
          'enable_custom_tiles_per_page' => '',
          'tiles_per_page' => 15,
          'click_through_url' => self::CLICK_THROUGH_URL_1,
          'polling_frequency' => '30',
          'parent_page_secret_key' => '',
          'enable_scrolling' => '0',
          'text_tile_background' => 'ffffff',
          'text_tile_font_color' => self::TEXT_TILE_FONT_COLOR,
          'text_tile_font_size' => 24,
          'text_tile_user_name_font_color' => self::TEXT_TILE_USER_NAME_FONT_COLOR,
          'text_tile_user_name_font_size' => 18,
          'text_tile_user_handle_font_color' => self::TEXT_TILE_USER_NAME_FONT_COLOR,
          'text_tile_user_handle_font_size' => 18,
          'margin' => 0,
          'max_tile_width' => 365,
          'widget_background' => "",
          'widget_loading_image' => self::WIDGET_LOADING_IMAGE,
        ];
        $style = $style + $defaultStyle;
        $config['tile_options'] = array_merge($defaultConfig['tile_options'], $config['tile_options']);
        $config['lightbox'] = array_merge($defaultConfig['lightbox'], $config['lightbox']);
        break;

      case self::STYLE_BASE_CAROUSEL:
      case self::STYLE_HORIZONTAL_FUILD:
        $defaultStyle = [
          'enable_custom_tiles_per_page' => '',
          'tiles_per_page' => 15,
          'click_through_url' => self::CLICK_THROUGH_URL_1,
          'polling_frequency' => '30',
          'parent_page_secret_key' => '',
          'widget_background' => "",
          'text_tile_background' => '',
          'text_tile_font_color' => self::TEXT_TILE_FONT_COLOR,
          'text_tile_source_color' => self::TEXT_TILE_USER_NAME_FONT_COLOR,
          'margin' => 7,
          'widget_height' => 300,
          'widget_loading_image' => self::WIDGET_LOADING_IMAGE,
        ];
        $style = $style + $defaultStyle;
        $config['tile_options'] = array_merge($defaultConfig['tile_options'], $config['tile_options']);
        $config['lightbox'] = array_merge($defaultConfig['lightbox'], $config['lightbox']);
        break;

      case self::STYLE_BASE_BILLBOARD:
      case self::STYLE_CAROUSEL:
        $defaultStyle = [
          'navigation' => 'click',
          'click_through_url' => self::CLICK_THROUGH_URL_2,
          'polling_frequency' => '30',
          'widget_background' => "",
          'arrow_color' => "333333",
          'text_tile_user_name_font_color' => self::TEXT_TILE_FONT_COLOR,
          'text_tile_user_name_font_size' => 16,
          'text_tile_user_handle_font_color' => self::TEXT_TILE_FONT_COLOR,
          'text_tile_user_handle_font_size' => 13,
          'text_tile_background' => 'ffffff',
          'text_tile_font_color' => self::TEXT_TILE_FONT_COLOR,
          'text_tile_font_size' => 17,
          'image_tile_background' => 'ffffff',
          'image_tile_font_color' => 'ffffff',
          'image_tile_font_size' => 17,
          'width' => 970,
          'height' => 300,
          'margin' => 7,
          'rows' => 10,
          'columns' => 2,
          'tileWidth' => 300,
          'tileHeight' => 300,
          'tiles_per_page' => 9,
          'widget_loading_image' => self::WIDGET_LOADING_IMAGE,
        ];
        $style = $style + $defaultStyle;
        break;

      case self::STYLE_BASE_FEED:
      case self::STYLE_SCROLL:
        $defaultStyle = [
          'navigation' => 'click',
          'click_through_url' => self::CLICK_THROUGH_URL_2,
          'polling_frequency' => '30',
          'widget_background' => "ffffff",
          'arrow_color' => "000000",
          'text_tile_user_name_font_color' => self::TEXT_TILE_FONT_COLOR,
          'text_tile_user_name_font_size' => 16,
          'text_tile_user_handle_font_color' => self::TEXT_TILE_FONT_COLOR,
          'text_tile_user_handle_font_size' => 13,
          'text_tile_background' => 'ffffff',
          'text_tile_font_color' => self::TEXT_TILE_FONT_COLOR,
          'text_tile_font_size' => 18,
          'image_tile_background' => 'ffffff',
          'image_tile_font_color' => self::TEXT_TILE_FONT_COLOR,
          'image_tile_font_size' => 18,
          'width' => 970,
          'height' => 600,
          'margin' => 15,
          'rows' => 3,
          'columns' => 3,
          'tileWidth' => 253,
          'tileHeight' => 253,
          'tiles_per_page' => 9,
          'widget_loading_image' => self::WIDGET_LOADING_IMAGE,
        ];
        $style = $style + $defaultStyle;
        break;

      case self::STYLE_BASE_SLIDESHOW:
      case self::STYLE_SLIDESHOW:
        $defaultStyle = [
          'click_through_url' => self::CLICK_THROUGH_URL_2,
          'polling_frequency' => '30',
          'widget_background' => "ffffff",
          'arrow_color' => "333333",
          'text_tile_user_name_font_color' => self::TEXT_TILE_FONT_COLOR,
          'text_tile_user_name_font_size' => 24,
          'text_tile_user_handle_font_color' => self::TEXT_TILE_FONT_COLOR,
          'text_tile_user_handle_font_size' => 20,
          'text_tile_background' => 'ffffff',
          'text_tile_font_color' => self::TEXT_TILE_FONT_COLOR,
          'text_tile_font_size' => 24,
          'image_tile_background' => 'ffffff',
          'image_tile_font_color' => 'ffffff',
          'image_tile_font_size' => 24,
          'width' => 970,
          'height' => 600,
          'margin' => 0,
          'rows' => 1,
          'columns' => 1,
          'tileWidth' => 970,
          'tileHeight' => 600,
          'tiles_per_page' => 15,
          'widget_loading_image' => self::WIDGET_LOADING_IMAGE,
        ];
        $style = $style + $defaultStyle;
        break;

      default:
        break;
    }
    $this->style = $style;
    $this->config = $config;
  }

  /**
   * Create function.
   *
   * @return StacklaModel|Widget|false|mixed
   *   The result.
   */
  public function create() {
    $endpointCall = sprintf("%s", $this->endpoint);

    $this->initRequest();

    $this->setDefaultValue();

    $data = $this->toArray();

    $json = $this->request->sendPost($endpointCall, [], ['body' => json_encode($data)]);

    $this->fromJson($json);

    return $json === FALSE ? FALSE : $this;
  }

  /**
   * Duplicate function.
   *
   * @return mixed|bool|Widget
   *   The result.
   */
  public function duplicate() {
    $endpointCall = sprintf("%s/%s?action=clone", $this->endpoint, $this->id);

    $this->initRequest();

    $data = $this->toArray(TRUE);

    $json = $this->request->sendPost($endpointCall, [], ['body' => json_encode($data)]);

    $class = get_class($this);
    $widget = new $class($this->configs, $json);

    return $json === FALSE ? FALSE : $widget;
  }

  /**
   * Derive function.
   *
   * @param int $filter_id
   *   The filter id.
   * @param string $name
   *   The name.
   *
   * @return mixed|bool|Widget
   *   The result.
   */
  public function derive(int $filter_id, string $name) {
    $endpointCall = sprintf("%s/%s?action=derive", $this->endpoint, $this->id);

    $this->initRequest();

    $data = [
      'filter_id' => $filter_id,
      'style' => [
        'name' => $name,
      ],
    ];

    $json = $this->request->sendPost($endpointCall, [], ['body' => json_encode($data)]);

    $class = get_class($this);
    $widget = new $class($this->configs, $json);

    return $json === FALSE ? FALSE : $widget;
  }

  /**
   * Update function.
   *
   * @param bool $force
   *   The force.
   *
   * @return StacklaModel|Widget|false|mixed
   *   The result.
   *
   * @throws \Exception
   */
  public function update(bool $force = FALSE) {
    if ($this->isPlaceholder && !$force) {
      throw new \Exception(
        "This is placeholder object, it doesn't have a uptodate data. If you still want to update this object with provided property(ies), you can pass 'true' value to the first parameter of this method"
      );
    }

    $endpointCall = sprintf("%s/%s", $this->endpoint, $this->id);

    $this->initRequest();

    $this->setDefaultValue();

    $data = $this->toArray(TRUE);

    $json = $this->request->sendPut($endpointCall, [], ['body' => json_encode($data)]);

    $this->fromJson($json);

    return $json === FALSE ? FALSE : $this;
  }

}
