<?php

namespace Drupal\stackla_widget\Controller;

use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Controller\ControllerBase;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\Core\Messenger\Messenger;
use Drupal\Core\Url;
use Drupal\stackla_widget\Api\Credentials;
use Drupal\stackla_widget\Service\StacklaService;
use Symfony\Component\HttpFoundation\RedirectResponse;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\RequestStack;

/**
 * Provides Oauth authentication action.
 */
class StacklaController extends ControllerBase {

  /**
   * Stackla service.
   *
   * @var \Drupal\stackla_widget\Service\StacklaService
   */
  protected $stackla;

  /**
   * The config factory service.
   *
   * @var \Drupal\Core\Config\ImmutableConfig
   */
  protected $config;

  /**
   * Core messenger.
   *
   * @var \Drupal\Core\Messenger\Messenger
   */
  protected $messenger;

  /**
   * Logger Factory.
   *
   * @var \Drupal\Core\Logger\LoggerChannelFactory
   */
  protected $loggerFactory;

  /**
   * Request stack.
   *
   * @var \Symfony\Component\HttpFoundation\RequestStack
   */
  protected $requestStack;

  /**
   * The StacklaController constructor.
   *
   * @param \Drupal\stackla_widget\Service\StacklaService $stackla
   *   The stackla service.
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   The config factory.
   * @param \Drupal\Core\Messenger\Messenger $messenger
   *   The messenger service.
   * @param \Drupal\Core\Logger\LoggerChannelFactoryInterface $logger_factory
   *   The logger factory.
   * @param \Symfony\Component\HttpFoundation\RequestStack $request_stack
   *   The request stack.
   */
  public function __construct(StacklaService $stackla, ConfigFactoryInterface $config_factory, Messenger $messenger, LoggerChannelFactoryInterface $logger_factory, RequestStack $request_stack) {
    $this->stackla = $stackla;
    $this->config = $config_factory->get('stackla_widget.settings');
    $this->messenger = $messenger;
    $this->loggerFactory = $logger_factory->get('stackla_widget');
    $this->requestStack = $request_stack;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('stackla_widget.stackla_service'),
      $container->get('config.factory'),
      $container->get('messenger'),
      $container->get('logger.factory'),
      $container->get('request_stack')
    );
  }

  /**
   * Oauth authentication.
   */
  public function stacklaOauth() {
    // Clear any other messages.
    $this->messenger->deleteAll();
    $request = $this->requestStack->getCurrentRequest()->query->all();
    $stackla_debug_mode = $this->config->get('debug_mode');
    if (isset($request['code'])) {
      $access_code = $request['code'];
      $client_id = $this->config->get('client_id');
      $host = $this->config->get('stack_oauth_host');
      $stack = $this->config->get('stack_name');
      $client_secret = $this->config->get('client_secret');
      $callback = Url::fromRoute('stackla_widget.stackla_oauth', [], ['absolute' => TRUE])
        ->toString();
      if ($stackla_debug_mode == 1) {
        $this->loggerFactory
          ->notice('Sending token request: host: @host, stack: @stack client id: @client_id, client secret: @client_secret, access code: @access_code, callback: @callback', [
            '@stack' => $stack,
            '@host' => $host,
            '@client_id' => $client_id,
            '@client_secret' => $client_secret,
            '@access_code' => $access_code,
            '@callback' => $callback,
          ]);
      }
      $credentials = new Credentials($this->config, $host, '', $stack, Credentials::TYPE_OAUTH2);
      $response = $credentials->generateToken($client_id, $client_secret, $access_code, $callback);
      if ($response === FALSE) {
        $this->messenger->addError('Failed to authorize with Stackla.');
      }
      else {
        $this->stackla->stacklaSetAccessToken($credentials->token);
        $this->messenger->addStatus(
          'Successfully authorized with Stackla');
      }
    }
    elseif (isset($request['error']) && isset($request['error_description'])) {
      $this->messenger->addError($request['error_description']);
    }
    return new RedirectResponse(Url::fromRoute('stackla_widget.stackla_settings_form')->setAbsolute()->toString());
  }

}
