<?php

namespace Drupal\stackla_widget\Form;

use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Form\ConfigFormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Messenger\Messenger;
use Drupal\Core\Url;
use Drupal\stackla_widget\Service\StacklaService;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * The AdminForm settings.
 *
 * @package Drupal\stackla_widget\Form
 */
class AdminForm extends ConfigFormBase {

  /**
   * Stackla service.
   *
   * @var \Drupal\stackla_widget\Service\StacklaService
   */
  protected $stackla;

  /**
   * Core messenger.
   *
   * @var \Drupal\Core\Messenger\Messenger
   */
  protected $messenger;

  /**
   * KeepeekdamConfig constructor.
   *
   * {@inheritdoc}
   */
  public function __construct(ConfigFactoryInterface $config_factory, Messenger $messenger, StacklaService $stackla) {
    parent::__construct($config_factory);
    $this->messenger = $messenger;
    $this->stackla = $stackla;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('config.factory'),
      $container->get('messenger'),
      $container->get('stackla_widget.stackla_service'),
    );
  }

  /**
   * {@inheritdoc}
   */
  protected function getEditableConfigNames() {
    return [
      'stackla_widget.settings',
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'stackla_widget_form';
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {

    $config = $this->config('stackla_widget.settings');

    $oauth_link = Url::fromRoute('stackla_widget.stackla_oauth', [], ['absolute' => TRUE])->toString();

    $form['for-stackla'] = [
      "#type" => 'fieldset',
      "#title" => $this->t('Step 1: Configure Stackla'),
      "#description" => $this->t("<p>Before you can configure Drupal, you need to first configure Stackla.</p>
        <p>Copy the 'URL' below and paste it in the relevant field in the Drupal plugin configuration screen in Stackla.</p>"),
    ];

    $form['for-stackla']['stackla_oauth'] = [
      '#title' => $this->t('Callback URL'),
      '#type' => 'textfield',
      '#value' => $oauth_link,
      '#disabled' => TRUE,
      '#description' => $this->t("Please copy + paste this URL into the 'Callback URL' field in the Drupal plugin configuration screen in Stackla."),
    ];

    $form['from-stackla'] = [
      '#type' => 'fieldset',
      '#title' => $this->t('Step 2: Configure Drupal'),
      '#description' => $this->t('You can retrieve these details from the Drupal plugin configuration screen.'),
    ];

    $form['from-stackla']['stack_name'] = [
      '#title' => $this->t('Stack Shortname'),
      '#type' => 'textfield',
      '#required' => TRUE,
      '#default_value' => $config->get('stack_name'),
    ];

    $form['from-stackla']['client_id'] = [
      '#title' => $this->t('Client ID'),
      '#type' => 'textfield',
      '#required' => TRUE,
      '#default_value' => $config->get('client_id'),
    ];

    $form['from-stackla']['client_secret'] = [
      '#title' => $this->t('Client Secret'),
      '#type' => 'textfield',
      '#required' => TRUE,
      '#default_value' => $config->get('client_secret'),
    ];

    $form['from-stackla']['refresh_interval'] = [
      '#title' => $this->t('Refresh interval'),
      '#type' => 'textfield',
      '#required' => TRUE,
      '#default_value' => $config->get('refresh_interval'),
      '#description' => $this->t('Value in seconds. -1 interval means run on every cron.'),
    ];

    $form['auth'] = [
      '#type' => 'fieldset',
      '#title' => $this->t('Step 3: Authorize'),
      '#description' => $this->t("You need to authorize Drupal to access your Stackla account. Click 'Authorize' below."),
    ];
    if ($this->stackla->stacklaCheckRequirements()) {
      if (!($this->stackla->stacklaIsAuthenticated())) {

        // Add authentication link.
        $form['auth']['authenticate'] = [
          '#type' => 'submit',
          '#value' => $this->t('Authorize'),
          '#op' => 'authenticate',
          '#submit' => ['::stacklaAuthenticate'],
        ];
      }
      else {
        $form['auth']['revoke_authentication'] = [
          '#type' => 'submit',
          '#value' => $this->t('Revoke authorization'),
          '#op' => 'authenticate',
          '#submit' => ['::stacklaAdminRevokeAuthentication'],
        ];

        $form['auth']['reauthenticate'] = [
          '#type' => 'submit',
          '#value' => $this->t('Reauthorize'),
          '#op' => 'authenticate',
          '#submit' => ['::stacklaAdminAuthenticate'],
        ];
      }
    }

    $form['proxy'] = [
      '#type' => 'fieldset',
      '#title' => $this->t('Proxy configuration'),
    ];

    $form['proxy']['proxy_status'] = [
      '#type' => 'radios',
      '#title' => $this->t('Proxy status'),
      '#options' => [
        1 => $this->t('Enable'),
        0 => $this->t('Disable'),
      ],
      '#default_value' => $config->get('proxy_status'),
    ];

    $form['proxy']['proxy_url'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Proxy URL'),
      '#default_value' => $config->get('proxy_url'),
      '#states' => [
        'visible' => [
          ':input[name="proxy_status"]' => ['value' => 1],
        ],
        'required' => [
          ':input[name="proxy_status"]' => ['value' => 1],
        ],
      ],
    ];

    $form['debug'] = [
      '#type' => 'fieldset',
      '#title' => $this->t('Debug'),
    ];

    $form['debug']['debug_mode'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Enable debug mode'),
      '#description' => $this->t('Will display and log verbose debug information. This should be disabled in production.'),
      '#default_value' => $config->get('debug_mode'),
    ];

    return parent::buildForm($form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    $this->config('stackla_widget.settings')
      ->set('stack_name', $form_state->getValue('stack_name'))
      ->set('client_id', $form_state->getValue('client_id'))
      ->set('client_secret', $form_state->getValue('client_secret'))
      ->set('refresh_interval', $form_state->getValue('refresh_interval'))
      ->set('proxy_status', $form_state->getValue('proxy_status'))
      ->set('proxy_url', $form_state->getValue('proxy_url'))
      ->set('debug_mode', $form_state->getValue('debug_mode'))
      ->save();
  }

  /**
   * Stackla authentication.
   *
   * @param array $form
   *   The form object.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The form_state object.
   */
  public function stacklaAuthenticate(array &$form, FormStateInterface &$form_state) {
    if ($this->stackla->stacklaCheckRequirements()) {
      $url = $this->stackla->stacklaAuthenticate();
      $form_state->setResponse($url);
    }

    return FALSE;
  }

  /**
   * Stackla authentication.
   *
   * @param array $form
   *   The form object.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The form_state object.
   */
  public function stacklaAdminAuthenticate(array &$form, FormStateInterface &$form_state) {
    $this->stacklaAuthenticate($form, $form_state);
  }

  /**
   * Revoke Stackla authentication.
   *
   * @param array $form
   *   The form object.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The form_state object.
   */
  public function stacklaAdminRevokeAuthentication(array &$form, FormStateInterface &$form_state) {
    if ($this->stackla->stacklaCheckRequirements()) {
      $this->stackla->stacklaRevokeAuthentication();
    }

    return FALSE;
  }

}
