<?php

namespace Drupal\stackla_widget\Service;

use Drupal\Core\Routing\TrustedRedirectResponse;
use Drupal\Core\Session\AccountProxyInterface;
use Drupal\Core\State\State;
use Drupal\Core\Cache\CacheBackendInterface;
use Drupal\Core\Url;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\stackla_widget\Api\Credentials;
use Drupal\stackla_widget\Api\Stack;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Drupal\Core\Messenger\Messenger;

/**
 * The Service Stackla.
 *
 * @package Drupal\stackla_widget\Service
 */
class StacklaService {

  /**
   * Permission adminstration Stackla.
   */
  const STACKLA_PERMISSION_ADMIN = 'administer stackla';

  /**
   * Permission use Stackla.
   */
  const STACKLA_PERMISSION_USE = 'use stackla';

  /**
   * The config factory service.
   *
   * @var \Drupal\Core\Config\ImmutableConfig
   */
  private $config;

  /**
   * Logger Factory.
   *
   * @var \Drupal\Core\Logger\LoggerChannelFactory
   */
  protected $loggerFactory;

  /**
   * The core messenger.
   *
   * @var \Drupal\Core\Messenger\Messenger
   */
  protected $messenger;

  /**
   * The current logged user.
   *
   * @var \Drupal\Core\Session\AccountProxyInterface
   */
  protected $currentUser;

  /**
   * The State service.
   *
   * @var \Drupal\Core\State\State
   */
  protected $state;

  /**
   * The cache factory service.
   *
   * @var \Drupal\Core\Cache\CacheBackendInterface
   */
  protected $cache;

  /**
   * Client constructor.
   *
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   The config factory.
   * @param \Drupal\Core\Logger\LoggerChannelFactoryInterface $logger_factory
   *   The logger factory.
   * @param \Drupal\Core\Messenger\Messenger $messenger
   *   The messenger.
   * @param \Drupal\Core\Session\AccountProxyInterface $current_user
   *   The current user connected.
   * @param \Drupal\Core\State\State $state
   *   The State service.
   * @param \Drupal\Core\Cache\CacheBackendInterface $cache
   *   The cache factory.
   */
  public function __construct(ConfigFactoryInterface $config_factory, LoggerChannelFactoryInterface $logger_factory, Messenger $messenger, AccountProxyInterface $current_user = NULL, State $state, CacheBackendInterface $cache) {
    $this->config = $config_factory->get('stackla_widget.settings');
    $this->loggerFactory = $logger_factory->get('stackla_widget');
    $this->messenger = $messenger;
    $this->currentUser = $current_user;
    $this->state = $state;
    $this->cache = $cache;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('config.factory'),
      $container->get('logger.factory'),
      $container->get('messenger'),
      $container->get('current_user'),
      $container->get('state'),
      $container->get('cache.bootstrap')
    );
  }

  /**
   * Redirect user to Stackla authentication.
   *
   * @return \Symfony\Component\HttpFoundation\RedirectResponse
   *   The Redirection object.
   */
  public function stacklaAuthenticate() {
    $client_id = $this->config->get('client_id');
    $host = $this->config->get('stack_oauth_host');
    $stack = $this->config->get('stack_name');
    $callback = Url::fromRoute('stackla_widget.stackla_oauth', [], ['absolute' => TRUE])
      ->toString();
    $credentials = new Credentials($this->config, $host, '', $stack, Credentials::TYPE_OAUTH2);
    $access_uri = $credentials->getAccessUri($client_id, $callback);

    return new TrustedRedirectResponse(Url::fromUri($access_uri)->toString());
  }

  /**
   * Revoke user authorization to Stackla.
   */
  public function stacklaRevokeAuthentication() {
    // Clear any other messages.
    $this->messenger->deleteAll();

    $this->stacklaUnsetAccessToken();
    $this->messenger->addStatus('Successfully revoked Stackla authentication');
  }

  /**
   * Check requirements module.
   *
   * @return bool
   *   The status of requirements check.
   */
  public function stacklaCheckRequirements() {
    if ($this->config->get('stack_oauth_host') &&
      $this->config->get('stack_name') &&
      $this->config->get('client_id') &&
      $this->config->get('client_secret') &&
      $this->stacklaUserRights()
    ) {
      return TRUE;
    }
    return FALSE;
  }

  /**
   * Check Stacka authentication.
   *
   * @return bool
   *   TRUE if authenticated.
   */
  public function stacklaIsAuthenticated() {
    // First check permissions.
    if (!$this->stacklaUserRights() || !$this->stacklaGetAccessToken() || !$this->stacklaCheckRequirements()) {

      return FALSE;
    }
    $stack = $this->stacklaCreateStack();
    $widget = $stack->instance('widget');
    try {
      $widget->get(5, 1)->getResults();
    }
    catch (\Exception $ex) {
      // Some sort of fetch error. It doesn't really matter which.
    }

    return ($widget->getResponseCode() == 200 ? TRUE : FALSE);
  }

  /**
   * Retrieve stackla credential variables.
   *
   * @return array
   *   Stackla credentials.
   */
  public function stacklaGetVariables() {

    return [
      'stack_name' => $this->config->get('stack_name'),
      'oauth_host' => $this->config->get('stack_oauth_host'),
      'host_name' => $this->config->get('host_name'),
      'access_token' => $this->stacklaGetAccessToken(),
    ];
  }

  /**
   * Create stack .
   *
   * @return object
   *   Return a Stack class.
   */
  public function stacklaCreateStack() {

    $stack = &drupal_static(__FUNCTION__);

    if (!isset($stack)) {
      // Load the SDK.
      $stackla_variables = $this->stacklaGetVariables();

      if ($this->config->get('stackla_debug_mode') == 1) {
        $this->loggerFactory->notice(
          'Attempting to create stack. Values: <pre>@values</pre>',
          [
            '@values' => print_r($stackla_variables, TRUE),
          ]
        );
      }
      // Stackla stack name.
      $stack_name = $stackla_variables['stack_name'];

      // Stackla Credentials details.
      $access_host = $stackla_variables['oauth_host'];
      // OAuth2 access_token.
      $access_token = $stackla_variables['access_token'];
      $credentials = new Credentials($this->config, $access_host, $access_token, $stack_name, Credentials::TYPE_OAUTH2);

      // Stackla API configs.
      $stack_host = $stackla_variables['host_name'];
      $stack = new Stack($credentials, $stack_host, $stack_name);
    }
    return $stack;
  }

  /**
   * Check if user has right access to Stackla.
   *
   * @return bool
   *   The status of user rights check.
   */
  public function stacklaUserRights() {
    // Check that the current user has the correct permissions set up.
    $user_stackla_admin = $this->currentUser
      ->hasPermission(self::STACKLA_PERMISSION_ADMIN);

    $user_stackla_use = $this->currentUser
      ->hasPermission(self::STACKLA_PERMISSION_USE);

    return (!$user_stackla_admin && !$user_stackla_use) ? FALSE : TRUE;
  }

  /**
   * Retrieve Stackla access token.
   *
   * @return string
   *   The access token string or FALSE if not present.
   */
  public function stacklaGetAccessToken() {

    $token = $this->state->get('stackla_access_token');
    if (!empty($token)) {

      return $token;
    }
    else {

      return FALSE;
    }

  }

  /**
   * Set Stackla access token.
   *
   * @param string $token
   *   The token string to set.
   *
   * @return bool
   *   TRUE if successful.
   */
  public function stacklaSetAccessToken(string $token) {
    if (!empty($token)) {
      $this->state->set('stackla_access_token', $token);

      return TRUE;
    }
    else {

      return FALSE;
    }
  }

  /**
   * Unset Stackla access token.
   *
   * @return bool
   *   TRUE if successful.
   */
  public function stacklaUnsetAccessToken() {
    $this->state->delete('stackla_access_token');

    return TRUE;
  }

  /**
   * Get list of stackla widgets for select options.
   *
   * @return array
   *   The list of options.
   */
  public function getWidgetsOptions() {
    $final = [];
    $cache = $this->cache->get('stackla_widget:widgets_list');
    if (!empty($cache)) {
      $options = $cache->data;
    }
    if (empty($options)) {
      $options = $this->getWidgetsList();
      $this->cache->set('stackla_widget:widgets_list', $options);
    }
    foreach ($options as $option) {
      $label = $option->getName();
      $id = $option->toArray()['id'];
      $final[$id] = $label;
    }

    return $final;
  }

  /**
   * Get list of stackla widgets for select options.
   *
   * @param int $page
   *   Number of page.
   * @param int $count
   *   Number of elements per page.
   *
   * @return array
   *   The list of options.
   */
  public function getWidgetsList(int $page = 1, int $count = 500) {
    $stack = $this->stacklaCreateStack();
    $widgets = $stack->instance('widget');

    return $widgets->get($count, $page)->getResults();
  }

  /**
   * Get Widget Embed code by ID.
   *
   * @param int|string $id
   *   The id of Stackla widget.
   *
   * @return mixed
   *   The embed code to display.
   */
  public function getIframeById($id) {
    $stack = $this->stacklaCreateStack();
    $widgets = $stack->instance('widget');
    $widget = $widgets->getById($id);

    return $widget->toArray()['embed_code'];
  }

}
